/*
 * ts_ttgo.c
 * 
 * Input driver for the GO touchscreen. 
 * Generates a KEY_ESC key the touchscreen is touched. This key si by default assigned to
 * the QUIT command. 
 *
 * Juan-Jose De Orus - TomTom NV - 2004-08-26
 */

#include "../config.h"

#include "ts_ttgo.h"
#include "input.h"
#include "../osdep/keycodes.h"

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <barcelona/Barc_ts.h>

#ifndef TS_DEV
#define TS_DEV "/dev/ts"
#endif

/**
 * Removes any pending input event from the touchscreen device.
 *
 * @param fd File decriptor of the tochscreen device.
 * @param ms Milliseconds to wait for detecting input activity in the device. Pass zero
 *           to prevent any waiting.
 *
 * @return Number of pending events removed. A negative value (error code) in case of I/O error.
 */
static int flush_events(int fd,long ms) {
  TS_EVENT event;
  fd_set fdset;
  struct timeval tv;
  int ret;
  int cnt;
  
  FD_ZERO(&fdset);
  FD_SET(fd,&fdset);
  tv.tv_sec  = 0L;
  tv.tv_usec = ms * 1000;
  
  for ( cnt = 0; (ret = select(FD_SETSIZE,&fdset,NULL,NULL,&tv)) > 0 /*&& FD_ISSET(fd,&fdset)*/; ++cnt )
    if ( read(fd,&event,sizeof(TS_EVENT)) != sizeof(TS_EVENT) )
      break;

  return ret < 0 ? ret : cnt;
}

/**
 * Initialize the touchscreen device.
 *
 * @param dev Touchscreen device name.
 *
 * @return file descriptor of the open device or -1 if error. 
 */ 
int mp_input_ts_init(const char* dev) {
  int fd;
  
  if( (fd = open(dev ? dev : TS_DEV,O_RDONLY)) < 0 )
    return -1;
 
  // Remove events entered before mplayer initialisation...
  if ( flush_events(fd,0) < 0 ) {
    close(fd);
    return -1;
  }

  return fd;
}

/**
 * Waits for an event to be available from the device. Returns 'KEY_ESC' which will make
 * mplayer to quit. 

  STK: add more complex behaviour
 */
int mp_input_ts_read(int fd) 
{
  TS_EVENT event;

  if( read( fd, &event, sizeof( TS_EVENT ) ) == sizeof( TS_EVENT ) )
  {
   //printf( "Touchscreen: x %d, y %d, press %d\n", event.x, event.y, event.pressure );
   
   if( !event.pressure )
    if( event.y > 120 )
     return KEY_ESC;
    else
     if( event.x < 100 )
      return '/'; // Decrease volume
     else
      if( event.x > 220 )
       return '*'; // Increase volume
      else
       return 'm'; // mute
  }
  else
  {
   return  MP_INPUT_NOTHING;
  }
}

/**
 * Closes the touchscreen device.
 */
void mp_input_ts_close(int fd) {
  if ( fd != -1 ) {
    // Remove pending events before closing...
    flush_events(fd,100);
    close(fd);
  }
}
