
#include <sys/ioctl.h>
#include <linux/soundcard.h>
#include <unistd.h>
#include <math.h>
#include <errno.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <stdio.h>
#include <memory.h>
#include <assert.h>

#include "barcelona/Barc_snd.h"

#include "CoolSound.h"

#define TRACE_SOUND 1

#define SND_DEVICE_NAME "/dev/barc-snd"

#define min( a, b ) ( ( ( a ) < ( b ) ) ? ( a ): ( b ) )
#define max( a, b ) ( ( ( a ) > ( b ) ) ? ( a ): ( b ) )

// chunksize should _not_ be set lower than 1024 until the sd card driver is fixed. 
enum { KChunkSize         = 1024,
       KChunkCount        = 3,
       KPrimaryBufferSize = ( KChunkSize  * 2 ) * KChunkCount
    };

enum { KVolumeMax     = 121,
       KVolumeMin     = 90,
       KDefaultVolume = 10
    };



static struct  
{
 int iFileDescriptor;
 
 struct         cs_buffer_description         iBufferDescription;
 struct         cs_primary_buffer_description iPrimaryBufferDescription;
 int                                          iBufferHandle;
 unsigned char*                               iBufferPointer;
 int                                          iPlaybackOffset;
 int                                          iPlaying;
 int                                          iVolume;
} iCoolSoundData;

static int iInitialised = 0;

static int GetChunkByteSize()
{
 return KChunkSize / 8 *
        iCoolSoundData.iBufferDescription.bits_per_sample *  
        iCoolSoundData.iBufferDescription.channels;
}

static int GetBufferSize()
{
 return iCoolSoundData.iBufferDescription.samples / 8 *
        iCoolSoundData.iBufferDescription.bits_per_sample *  
        iCoolSoundData.iBufferDescription.channels;
}

static int GetPos()
{
 return ioctl( iCoolSoundData.iFileDescriptor, 
               COOLSOUND_GETPOS, 
               iCoolSoundData.iBufferHandle );
}

int CoolSoundInit( int aSampleRate, int aChannels )
{
 assert( !iInitialised );

 int tmpResult = -1;

 memset( &iCoolSoundData, 0, sizeof( iCoolSoundData ) );
 iCoolSoundData.iVolume = KDefaultVolume;

 if( ( tmpResult = iCoolSoundData.iFileDescriptor = open( SND_DEVICE_NAME, O_RDWR ) ) < 0 )
  {
   printf( "Cool sound %s could not be opened (%d)\n", SND_DEVICE_NAME, iCoolSoundData.iFileDescriptor );
  }
 else
 {
  /* setup play and record buffer parameters */
  iCoolSoundData.iBufferDescription.frequency       = aSampleRate;
  iCoolSoundData.iBufferDescription.channels        = aChannels;
  iCoolSoundData.iBufferDescription.bits_per_sample = 16;
  iCoolSoundData.iBufferDescription.samples         = 16 * KChunkSize;
  iCoolSoundData.iBufferDescription.type            = CSB_TYPE_PLAYBACK;

  //printf("cs:9 get buffer\n");
  iCoolSoundData.iBufferHandle = ioctl( iCoolSoundData.iFileDescriptor, 
                                        COOLSOUND_GET_BUFFER, 
                                        &iCoolSoundData.iBufferDescription );

  ioctl( iCoolSoundData.iFileDescriptor, 
         COOLSOUND_SEL_BUFFER, 
         iCoolSoundData.iBufferHandle );
  iCoolSoundData.iBufferPointer = (unsigned char*)mmap( 0, 
                                                        GetBufferSize(), 
                                                        PROT_READ | PROT_WRITE, 
                                                        MAP_SHARED, 
                                                        iCoolSoundData.iFileDescriptor, 
                                                        0 );
  if( iCoolSoundData.iBufferPointer == MAP_FAILED )
  {
   printf("Cool sound MMap failed\n" );
  }
  else
  {
   assert( NULL != iCoolSoundData.iBufferPointer );
   
   memset( iCoolSoundData.iBufferPointer, 0, GetBufferSize() );

   // Primary buffer setup
   iCoolSoundData.iPrimaryBufferDescription.frequency       = aSampleRate;
   iCoolSoundData.iPrimaryBufferDescription.channels        = 2;
   iCoolSoundData.iPrimaryBufferDescription.bits_per_sample = 16;
   iCoolSoundData.iPrimaryBufferDescription.chunksize       = KChunkSize;
   iCoolSoundData.iPrimaryBufferDescription.chunks          = KChunkCount;
  
    //printf("cs:11 setup primary\n");    
    int error = 1;
    int count = 0;

    while( error && count < 20 )
    {
     error = ioctl( iCoolSoundData.iFileDescriptor, 
                    COOLSOUND_SETUP_PRIMARY, 
                    &iCoolSoundData.iPrimaryBufferDescription );
     printf("Cool sound Setting up primary buffer: error = %d\n", error );

     if( error )
     {
      count++;
      usleep( 1000 );
     }
    };

    tmpResult = ( error ) ? -1: 0;

    if( !tmpResult )
    { // Final setup
     struct cb_chpos cbp;

     cbp.buffer   = iCoolSoundData.iBufferHandle;
     cbp.position = 0;

     if( ( ioctl( iCoolSoundData.iFileDescriptor, 
                  COOLSOUND_SETPOS, 
                  &cbp ) ) == -1 )
      printf("Cool sound SETPOS failed\n");

     iCoolSoundData.iPlaybackOffset = GetPos();

     iInitialised = 1;
     
     CoolSoundSetVolume( iCoolSoundData.iVolume );
     //CoolSoundMute( 0 );

     printf( "Cool sound setup %s OK, file handle %d, buffer handle %d, address 0x%p, position %d, %dHz, %dch\n", 
       SND_DEVICE_NAME,
       iCoolSoundData.iFileDescriptor,
       iCoolSoundData.iBufferHandle,
       iCoolSoundData.iBufferPointer,
       iCoolSoundData.iPlaybackOffset,
       iCoolSoundData.iBufferDescription.frequency,
       iCoolSoundData.iBufferDescription.channels
       );     
    }
  }
 }

 return tmpResult;
}

void CoolSoundDone()
{
 if( iInitialised )
 {
  CoolSoundMute( 1 );
  CoolSoundStop();

  munmap( iCoolSoundData.iBufferPointer, GetBufferSize() );

  iCoolSoundData.iBufferHandle = ioctl( iCoolSoundData.iFileDescriptor, 
                                        COOLSOUND_PUT_BUFFER, 
                                        iCoolSoundData.iBufferHandle );

  close( iCoolSoundData.iFileDescriptor );

  iInitialised = 0;

  printf( "Cool sound done\n" );
 }
}

static int GetMaxPlaySize()
{
 int tmpDevPos = GetPos();

 if( tmpDevPos == iCoolSoundData.iPlaybackOffset )
 { // Buffer empty
  return GetBufferSize();
 }
 else
  if( tmpDevPos > iCoolSoundData.iPlaybackOffset )
  {
   return tmpDevPos - iCoolSoundData.iPlaybackOffset;
  }
  else
  {
   return GetBufferSize() - iCoolSoundData.iPlaybackOffset;
  }
}

void CoolSoundPause( int aPause )
{
 if( !aPause )
 {
  if( !iCoolSoundData.iPlaying ) // Only if we do not play
   if( ( ioctl( iCoolSoundData.iFileDescriptor, 
                COOLSOUND_RUN_BUFFER, 
                iCoolSoundData.iBufferHandle ) ) == -1 )
   {
    printf( "Cool sound Run buffer failed\n" );
   }
   else
   {
    iCoolSoundData.iPlaying = 1;
    printf( "Cool sound Run buffer success\n" );
   }
 }
 else
 {
  if( iCoolSoundData.iPlaying ) // Only if we are playing
   if( ( ioctl( iCoolSoundData.iFileDescriptor, 
                COOLSOUND_STOP_BUFFER, 
                iCoolSoundData.iBufferHandle ) ) == -1 )
   {
    printf( "Cool sound Stop buffer failed\n" );
   }
   else
   {
    iCoolSoundData.iPlaying = 0;
    printf( "Cool sound Stop buffer success\n" );
   }
 }
}

int CoolSoundPlay( char* aBuffer, int aSize )
{
 assert( iInitialised );
 assert( NULL != aBuffer );
 //assert( aSize );
 
 if( !aSize )
 {
  return 0;
 }

 int tmpResult = min( aSize, GetMaxPlaySize() ); 

 memcpy( iCoolSoundData.iBufferPointer + iCoolSoundData.iPlaybackOffset,
         aBuffer, tmpResult );

 iCoolSoundData.iPlaybackOffset += tmpResult;

 /*
 if( tmpResult < GetChunkByteSize() )
 {
  memset( iCoolSoundData.iBufferPointer + iCoolSoundData.iPlaybackOffset,
          0,
          GetChunkByteSize() - tmpResult );

  iCoolSoundData.iPlaybackOffset += ( GetChunkByteSize() - tmpResult );
 }*/

 if( !iCoolSoundData.iPlaying )
 {
  CoolSoundPause( 0 );
 }

 // Wrapping around
 iCoolSoundData.iPlaybackOffset %= GetBufferSize();

 //printf( "Cool sound Play %d of %d bytes, new pos %d, audio pos %d\n", tmpResult, aSize, iCoolSoundData.iPlaybackOffset, GetPos() );

 return tmpResult;
}

void CoolSoundStop()
{
 if( iInitialised )
 {
  CoolSoundPause( 1 );

  iCoolSoundData.iPlaying = 0;

  struct cb_chpos cbp;

  cbp.buffer   = iCoolSoundData.iBufferHandle;
  cbp.position = 0;

  if( ( ioctl( iCoolSoundData.iFileDescriptor, 
               COOLSOUND_SETPOS, 
               &cbp ) ) == -1 )
   printf("Cool sound SETPOS failed\n");

  iCoolSoundData.iPlaybackOffset = 0;
 
  memset( iCoolSoundData.iBufferPointer, 0, GetBufferSize() );

  printf( "Cool sound stopped\n" );
 }
}

void CoolSoundStopWhenEmpty()
{
 if( iInitialised )
 {
  struct cb_qstop cqs;

  cqs.buffer        = iCoolSoundData.iBufferHandle;
  cqs.stop_position = iCoolSoundData.iPlaybackOffset;
 
  //printf("cs:6 queue stop \n");
  if( ioctl( iCoolSoundData.iFileDescriptor, 
             COOLSOUND_QUEUE_STOP, 
             &cqs ) != -1 ) 
  {
   printf( "Cool sound stop when pos %d\n", iCoolSoundData.iPlaybackOffset );
   iCoolSoundData.iPlaying = 0;
  }
 }
}

static int GetBusySize( int* aDevPos )
{
 *aDevPos = GetPos();

 if( iCoolSoundData.iPlaying )
 { 
  if( *aDevPos < iCoolSoundData.iPlaybackOffset )
  {
   return iCoolSoundData.iPlaybackOffset - *aDevPos;
  }
  else
  {
   return GetBufferSize() - *aDevPos + iCoolSoundData.iPlaybackOffset;
  }
 }
 else
  return 0;
}

int CoolSoundGetFreeSpace()
{
 //assert( iInitialised );

 int tmpDevPos   = 0;
 int tmpBusySize = GetBusySize( &tmpDevPos ); 
 int tmpFreeSize = GetBufferSize() - tmpBusySize;
 int tmpMinSize  = GetChunkByteSize() * ( KChunkCount + 1 );

 //printf( "Cool sound buf size %d, busy %d, Free %d, buf pos %d, audio pos %d\n", GetBufferSize(), tmpBusySize, tmpFreeSize, iCoolSoundData.iPlaybackOffset, tmpDevPos );


 return ( ( tmpFreeSize <= tmpMinSize ) ? 0: tmpFreeSize );
}

float CoolSoundGetPlayBackTime()
{
 //assert( iInitialised );

 int tmpDevPos   = 0;
 int tmpBusySize = GetBusySize( &tmpDevPos );

/* return 8.0 * ( float )GetBufferSize() / 
        ( float )iCoolSoundData.iBufferDescription.channels /
        ( float )iCoolSoundData.iBufferDescription.bits_per_sample /
        ( float )iCoolSoundData.iBufferDescription.frequency;*/

/* return ( float )tmpBusySize /        
        ( float )iCoolSoundData.iBufferDescription.frequency;*/

 if( tmpBusySize )
 {
  return 8.0 * ( ( float )tmpBusySize + GetChunkByteSize() * KChunkCount ) / 
        ( float )iCoolSoundData.iBufferDescription.channels /
        ( float )iCoolSoundData.iBufferDescription.bits_per_sample /
        ( float )iCoolSoundData.iBufferDescription.frequency;
 }
 else
 {
  return 0;
 }
}

int CoolSoundGetVolume()
{
 //assert( iInitialised );

 return iCoolSoundData.iVolume;
}

void CoolSoundMute( int aMute )
{
 if( iInitialised )
 {
  if( aMute )
  {
   ioctl( iCoolSoundData.iFileDescriptor, COOLSOUND_MUTE_INTERNAL, 0 );
   ioctl( iCoolSoundData.iFileDescriptor, COOLSOUND_MUTE_EXTERNAL, 0 );
   
   printf( "Cool sound muted\n" );
  }
  else
  {
   ioctl( iCoolSoundData.iFileDescriptor, COOLSOUND_UNMUTE_INTERNAL, 0 );
   ioctl( iCoolSoundData.iFileDescriptor, COOLSOUND_UNMUTE_EXTERNAL, 0 );
   
   printf( "Cool sound unmuted\n" );
  }
 }
}

void CoolSoundSetVolume( int aVolume )
{
 if( iInitialised )
 {
  if( aVolume < 0 )
   aVolume = KDefaultVolume;

  iCoolSoundData.iVolume = aVolume; 
 
  int tmpVolume = KVolumeMin + ( ( KVolumeMax - KVolumeMin ) * aVolume ) / 100; 

  int tmpResult = ioctl( 
                   iCoolSoundData.iFileDescriptor, 
                   COOLSOUND_SET_VOLUME, //BARC_SND_SET_VOLUME, 
                   tmpVolume );
   
  printf( "Cool sound volume set to %d (original %d), result %d\n", tmpVolume, aVolume, tmpResult );

  if( !aVolume )
  {
   CoolSoundMute( 1 );
  }
  else
  {
   CoolSoundMute( 0 );
  }
 }
}
