/*
 * ao_cool.c
 *
 * MPlayer sound driver for GO M100/300/500 using CoolSound driver.
 *
 * Juan-Jose De Orus - TomTom NV 2004-4
 */

#include <stdio.h>
#include <stdlib.h>

#include "../config.h"
#include "../mixer.h"
#include "../mp_msg.h"

#include "afmt.h"

#include "audio_out.h"
#include "audio_out_internal.h"

#include "CoolSound.h"
#include "TTGOWatchdog.h"

#define DEFAULT_VOLUME_FILE_NAME "/mnt/sdcard/bin/defvolume.dat"

static ao_info_t info = 
{
  "TomTom GO CoolSound audio output",
  "cool",
  "JJO and STK",
  ""
};

LIBAO_EXTERN(cool)


/**
 * Sets/gets/queries special features/parameters. 
 * 
 * @param cmd Control command (see list below).
 * @param arg Pointer to command argument, if any. 
 *
 * Supported commands:
 *
 * AOCONTROL_SET_DEVICE   Selects the audio device given in the argument.
 * AOCONTROL_GET_DEVICE   Returns the current audio device. The argument must
 *                        be a char** to hold the address of the internal 
 *                        name string.
 * AOCONTROL_QUERY_FORMAT Does nothing.
 * AOCONTROL_GET_VOLUME   Returns the current volume.
 * AOCONTROL_SET_VOLUME   Set the new volume.
 */                         
static int control(int cmd,void *arg) {
  switch( cmd ) 
  {
   case AOCONTROL_SET_DEVICE:
    return CONTROL_OK;
	
   case AOCONTROL_GET_DEVICE:
    *(char**)arg = "cool device";
    return CONTROL_OK;
	
   case AOCONTROL_QUERY_FORMAT:
    // TODO: Not used?? Should check if the given argument refers
    // to a supported format, and return a boolean with the result.
    mp_msg(MSGT_AO,MSGL_V,"[cool] Audio_control: querying format: %x)\n",*((int*)arg));
    return CONTROL_TRUE;
	
   case AOCONTROL_GET_VOLUME:    
      {
       ao_control_vol_t *vol = (ao_control_vol_t *)arg;
       vol->right = vol->left = CoolSoundGetVolume();
       mp_msg( MSGT_AO,MSGL_V, "[cool] Audio_control: querying volume: %d )\n", vol->right );

       return CONTROL_OK;
      }

   case AOCONTROL_SET_VOLUME: 
   {
    ao_control_vol_t *vol = (ao_control_vol_t *)arg;
    
    // Both channels are changed simultaneously. Choose the louder volume.
    int volume = vol->right >= vol->left ? vol->right : vol->left;

    // volume is in the rangle 0..100.
    CoolSoundSetVolume( volume );
  
    mp_msg( MSGT_AO,MSGL_V, "[cool] Audio_control: volume set to: %d\n", volume );

    return CONTROL_OK;
    }
  } 
  return CONTROL_UNKNOWN;
}

/**
 * Open & setup audio device.
 *
 * @param rate     Sample rate.
 * param  channels Number of channels.
 * @param format   Sample format.
 * @param flags    Not used.
 *
 * @return 1=success, 0=fail.
 */
static int init(int rate,int channels,int format,int flags) 
{
  int   error = 0;
  
  mp_msg(MSGT_AO,MSGL_V,
         "[cool] Init Audio_setup: %d Hz, %d chans, format name %s, format %d, flags %d\n",
         rate,
         channels,
         audio_out_format_name( format ),
         format,
         flags );

  // CoolSound only supports AFMT_S16_LE format.
  if ( format != AFMT_S16_LE ) {
    mp_msg(MSGT_AO,MSGL_ERR,
           "[cool] Audio_setup: format '%s' not supported\n", 
           audio_out_format_name(format));
    return 0;
  }

  // Only 2 audio channels.
  if ( channels > 2 ) {
    mp_msg(MSGT_AO,MSGL_ERR,
           "[cool] Audio_setup: Failed to set audio device to %d channels\n", 
           channels);
    return 0;
  }

  // Initializing watchdog  
  error = TTGOWatchdogInit();
  if( error )
  {
    mp_msg( MSGT_AO,MSGL_ERR,
            "[cool] Watchdog setup: Can not initialise GO watchdog, code %d\n", error );
    return 0;
  }
  else
  {
   mp_msg( MSGT_AO,MSGL_V, "Watchdog initialised\n" ); 
  }


  error = ( CoolSoundInit( rate, channels ) ? 0: 1 );

  if( error )
  {
   FILE* tmpVolumeFile = NULL;
   int   tmpVolume;
   
   if( NULL != ( tmpVolumeFile = fopen( DEFAULT_VOLUME_FILE_NAME, "rb" ) ) )
   {
    fscanf( tmpVolumeFile, "%d\n", &tmpVolume );
    fclose( tmpVolumeFile );
    CoolSoundSetVolume( tmpVolume );
   }
  }

  return error;
}

/**
 * Close audio device.
 *
 * @param immed if true, stops the device immediately. If false, waits until
 *              the last byte written to the device has been played.
 */
static void uninit(int immed) 
{
  const int tmpVolume     = CoolSoundGetVolume();
  FILE*     tmpVolumeFile = NULL;
  
  TTGOWatchdogDone();

  // Saving volume
  if( NULL != ( tmpVolumeFile = fopen( DEFAULT_VOLUME_FILE_NAME, "w+b" ) ) )
  {
   fprintf( tmpVolumeFile, "%d\n", tmpVolume );
   fclose( tmpVolumeFile );
  }

  CoolSoundDone();

  mp_msg(MSGT_AO,MSGL_V,
         "[cool] Audio_cleanup: Device closed)\n" );
}

/**
 * Stop playing and empty buffers (for seeking/pause).
 */
static void reset() 
{
 CoolSoundStop();
}

/**
 * Stop playing, keep buffers (for pause).
 */
static void audio_pause() 
{
 CoolSoundPause( 1 );
}

/**
 *  Resume playing, after audio_pause().
 */
static void audio_resume() 
{
 CoolSoundPause( 0 );
}

/**
 * Calculate how many bytes can be played without blocking (that is, the amount of
 * free space in the circular buffer). If the buffer is (nearly) full,  has to return 0!
 * If it never gives 0, MPlayer won't work!
 */
static int get_space() 
{
 /*int   tmpFreeSpace = CoolSoundGetFreeSpace();
 //float tmpToPlay    = CoolSoundGetPlayBackTime();

 printf( "========================== Free space %d\n", tmpFreeSpace );

 return tmpFreeSpace;*/

 return CoolSoundGetFreeSpace();
}

/**
 * Calculate delay in seconds between first and last sample in buffer.
 *
 * @return delay.
 */
static float get_delay() 
{
 /*//int   tmpFreeSpace = CoolSoundGetFreeSpace();
 float tmpToPlay    = CoolSoundGetPlayBackTime();

 printf( "========================== To play %f\n", tmpToPlay );

 return tmpToPlay;*/
 return CoolSoundGetPlayBackTime();
}

/**
 * Plays audio. 
 *
 * @param data Raw audio data.
 * @param len  Data size in bytes. 
 *
 * @return: number of bytes played.
 */
static int play(void* data,int len,int flags) 
{
 /*int tmpRes = CoolSoundPlay( data, len );

 printf( "========================== Play %d of %d\n", tmpRes, len );

 return tmpRes;*/
 TTGOWatchdogPing();
 
 return CoolSoundPlay( data, len );
}


