/*
 * ao_ttgo.c
 *
 * TomTomGO sound driver for MPlayer.
 *
 * Juan-Jose De Orus - TomTom NV 2004
 */

#include <stdio.h>
#include <stdlib.h>

#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>

#include "../config.h"
#include "../mp_msg.h"

#include "afmt.h"

#include "audio_out.h"
#include "audio_out_internal.h"

/**
 * The list below describes the ioctl commands implemented by the Barcelona 
 * sound driver.
 *
 * SNDCTL_DSP_STATUS      (uint)
 * SNDCTL_DSP_SETFMT      (long) Only AFMT_S16_LE.
 * SNDCTL_DSP_VOLUME      (int)
 * SNDCTL_DSP_CHANNELS    (long)
 * SNDCTL_DSP_STEREO      (long)
 * SNDCTL_DSP_SPEED       (long)
 * SNDCTL_DSP_GETFMTS     (long) Always (AFMT_S16_LE).
 * SNDCTL_DSP_GETBLKSIZE  (long for write buffers, int for read buffers).
 * SNDCTL_DSP_SETFRAGMENT (long for write buffers, int for read buffers).
 * SNDCTL_DSP_SYNC        (none)
 * SNDCTL_DSP_GETOSPACE   (audio_buf_info*)
 * SNDCTL_DSP_GETISPACE   (audio_buf_info*)
 * SNDCTL_DSP_RESET       (none) 
 * SNDCTL_DSP_NONBLOCK    (none)
 * SNDCTL_DSP_MUTE        (none)
 * SNDCTL_DSP_MUTEOFF     (none)
 * SNDCTL_DSP_ON          (none)
 * SNDCTL_DSP_OFF         (none)
 * SOUND_PCM_READ_CHANNELS (long)
 * SOUND_PCM_READ_RATE    (long)
 *
 * Unsupported commands:
 *
 * SNDCTL_DSP_POST          
 * SDCTL_DSP_SUBDIVIDE      
 * SNDCTL_DSP_GETCAPS       
 * SNDCTL_DSP_GETTRIGGER     
 * SNDCTL_DSP_SETTRIGGER    
 * SNDCTL_DSP_GETIPTR       
 * SNDCTL_DSP_GETOPTR       
 * SNDCTL_DSP_MAPINBUF      
 * SNDCTL_DSP_MAPOUTBUF     
 * SNDCTL_DSP_SETSYNCRO     
 * SNDCTL_DSP_SETDUPLEX     
 * SNDCTL_DSP_GETODELAY     
 * SNDCTL_DSP_GETCHANNELMASK 
 * SNDCTL_DSP_BIND_CHANNEL   
 * SNDCTL_DSP_SETSPDIF       
 * SNDCTL_DSP_GETSPDIF       
 * SNDCTL_DSP_PROFILE        
 * SOUND_PCM_READ_BITS       
 * SOUND_PCM_READ_FILTER     
 * SOUND_PCM_WRITE_FILTER   
 *  
 * Mixer:
 *
 * There is no mixer device in the TomTomGO, although some mixer commands are
 * supported by the driver. 
 */

#define FRAGMENTS        8   // Number of fragments to allocate.
#define FRAGMENT_SIZE    14  // Maximum fragment size allowed (in 2^n bytes => 2^14 = 16Kb). 
#define BYTES_PER_SAMPLE 2   // The GO only supports 16 bit samples. 
 
static ao_info_t info = {
  "TomTomGO ioctl audio output",
  "ttgo",
  "JJO",
  ""
};

LIBAO_EXTERN(ttgo)

static char *dsp=PATH_DEV_DSP;        // Defined in config.h as '/dev/dsp'.
static audio_buf_info zz;             // Info about the audio buffers.
static int audio_fd=-1;               // /dev/dsp file descriptor.
static int volume = 50;               // Default volume.

/**
 * Sets/gets/queries special features/parameters. 
 * 
 * @param cmd Control command (see list below).
 * @param arg Pointer to command argument, if any. 
 *
 * Supported commands:
 *
 * AOCONTROL_SET_DEVICE   Selects the audio device given in the argument.
 * AOCONTROL_GET_DEVICE   Returns the current audio device. The argument must
 *                        be a char** to hold the address of the internal 
 *                        name string.
 * AOCONTROL_QUERY_FORMAT Does nothing.
 * AOCONTROL_GET_VOLUME   Returns the current volume.
 * AOCONTROL_SET_VOLUME   Set the new volume.
 */                         
static int control(int cmd,void *arg) {
  switch( cmd ) {
  case AOCONTROL_SET_DEVICE:
    dsp=(char*)arg;
    return CONTROL_OK;
	
  case AOCONTROL_GET_DEVICE:
    *(char**)arg=dsp;
    return CONTROL_OK;
	
  case AOCONTROL_QUERY_FORMAT:
    // TODO: Not used?? Should check if the given argument refers
    // to a supported format, and return a boolean with the result.
    mp_msg(MSGT_AO,MSGL_V,"Audio control: querying format: %x)\n",*((int*)arg));
    return CONTROL_TRUE;
	
  case AOCONTROL_GET_VOLUME: {
    ao_control_vol_t *vol = (ao_control_vol_t *)arg;
    vol->right = vol->left = volume;
    mp_msg(MSGT_AO,MSGL_V,"Audio control: querying volume:  %d)\n",volume);
    return CONTROL_OK;
    }
  case AOCONTROL_SET_VOLUME: {
    ao_control_vol_t *vol = (ao_control_vol_t *)arg;
    
    // Both channels are changed simultaneously. Choose the louder volume.
    volume = vol->right >= vol->left ? vol->right : vol->left;
    if ( audio_fd != -1 )
      ioctl(audio_fd,SNDCTL_DSP_VOLUME,&volume);
    mp_msg(MSGT_AO,MSGL_V,"Audion control: volume set to: %d\n",volume);
    return CONTROL_OK;
    }
  } 
  return CONTROL_UNKNOWN;
}

/**
 * Open & setup audio device.
 *
 * @param rate     Sample rate.
 * param  channels Number of channels.
 * @param format   Sample format.
 * @param flags    Not used.
 *
 * @return 1=success, 0=fail.
 */
static int init(int rate,int channels,int format,int flags) {
  int frags;

  mp_msg(MSGT_AO,MSGL_V,
         "[ttgo] Audio setup: %d Hz  %d chans  %s\n",
         rate,
         channels,
         audio_out_format_name(format));
  mp_msg(MSGT_AO,MSGL_V,"Audio_setup: using '%s' dsp device\n", dsp);

  // TTGO only supports AFMT_S16_LE format.
  if ( format != AFMT_S16_LE ) {
    mp_msg(MSGT_AO,MSGL_ERR,
           "[ttgo] Audio_setup: format '%s' not supported\n", 
           audio_out_format_name(format));
    return 0;
  }

  // TTGO only supports 2 audio channels.
  if ( channels > 2 ) {
    mp_msg(MSGT_AO,MSGL_ERR,
           "[ttgo] Audio_setup: Failed to set audio device to %d channels\n", 
           channels);
    return 0;
  }

  if ( (audio_fd=open(dsp, O_WRONLY)) < 0 ) {
    mp_msg(MSGT_AO,MSGL_ERR,
           "[ttgo] Audio_setup: Can't open audio device %s: %s\n", 
           dsp, 
           strerror(errno));
    return 0;
  }
  
  if ( ioctl(audio_fd,SNDCTL_DSP_ON,NULL) < 0 ) {
    mp_msg(MSGT_AO,MSGL_ERR,"[ttgo] Audio_setup: Cannot enable device %s\n",dsp);
    return 0;
  }

  // Set the fragments.  
  frags = (FRAGMENTS << 16) + FRAGMENT_SIZE;    
  if ( ioctl(audio_fd,SNDCTL_DSP_SETFRAGMENT,&frags) < 0 ) {
    mp_msg(MSGT_AO,MSGL_ERR,
           "[ttgo] Audio_setup: Cannot set fragments: nr=%d, sz=%d\n",
           FRAGMENTS,
           1 << (FRAGMENT_SIZE));
    return 0;
  }

  if( ioctl(audio_fd, SNDCTL_DSP_GETOSPACE, &zz) < 0 ) {
    mp_msg(MSGT_AO,MSGL_ERR,
           "[ttgo] Audio_setup: Cannot get buffer info (SNDCTL_DSP_GETOSPACE)\n");
    return 0;
  } 

  mp_msg(MSGT_AO,MSGL_V,
         "[ttgo] Audio_setup: frags: %3d/%d  (%d bytes/frag)  free: %6d\n",
         zz.fragments, 
         zz.fragstotal, 
         zz.fragsize, 
         zz.bytes);
  
  if( ao_data.buffersize == -1 ) 
    ao_data.buffersize=zz.bytes;

  // Set audio data parameters.
  ao_data.format=format;
  ioctl(audio_fd, SNDCTL_DSP_SETFMT, &ao_data.format);
  mp_msg(MSGT_AO,MSGL_V,
         "[ttgo] Audio_setup: sample format: %s (requested: %s)\n",
         audio_out_format_name(ao_data.format), 
         audio_out_format_name(format));

  ao_data.channels = channels;
  ioctl(audio_fd, SNDCTL_DSP_CHANNELS, &ao_data.channels);
  mp_msg(MSGT_AO,MSGL_V,
         "[ttgo] Audio_setup: using %d channels (requested: %d)\n", 
         ao_data.channels, channels);
  
  ao_data.samplerate = rate;
  ioctl (audio_fd, SNDCTL_DSP_SPEED, &ao_data.samplerate);
  mp_msg(MSGT_AO,MSGL_V,
         "[ttgo] Audio_setup: using %d Hz samplerate (requested: %d)\n",
         ao_data.samplerate,rate);

  ao_data.outburst = zz.fragsize - (zz.fragsize % (ao_data.channels * BYTES_PER_SAMPLE));
  ao_data.bps      = ao_data.channels * BYTES_PER_SAMPLE * ao_data.samplerate;

  if ( ioctl(audio_fd,SNDCTL_DSP_MUTEOFF,NULL) < 0 )
    mp_msg(MSGT_AO,MSGL_WARN,"[ttgo] Audio_setup: Cannot unmute device %s\n",dsp);
      
  if ( ioctl(audio_fd,SNDCTL_DSP_VOLUME,&volume) < 0 ) 
    mp_msg(MSGT_AO,MSGL_WARN,"[ttgo] Audio_setup: Cannot change volume %d\n",volume);
  else
    mp_msg(MSGT_AO,MSGL_V,"[ttgo] Audio_setup: Using volume %d\n",volume);
  
  return 1;
}

/**
 * Close audio device.
 *
 * @param immed if true, stops the device immediately. If false, waits until
 *              the last byte written to the device has been played.
 */
static void uninit(int immed) {
  if ( audio_fd == -1 ) 
    return;

  ioctl(audio_fd, (immed ? SNDCTL_DSP_RESET : SNDCTL_DSP_SYNC) , NULL);
  close(audio_fd);
  audio_fd = -1;

  mp_msg(MSGT_AO,MSGL_V,
         "[ttgo] Audio_cleanup: dsp device %s closed)\n",dsp);
}

/**
 * Stop playing and empty buffers (for seeking/pause).
 */
static void reset() {
  uninit(1);
  if ( (audio_fd=open(dsp, O_WRONLY)) < 0 ) {
    mp_msg(MSGT_AO,MSGL_ERR,
           "[ttgo] Audio_reset: Can't reset audio device %s: %s\n", 
           dsp, 
           strerror(errno));
    return;
  }
 
  ioctl(audio_fd, SNDCTL_DSP_SETFMT, &ao_data.format);
  ioctl(audio_fd, SNDCTL_DSP_CHANNELS, &ao_data.channels);
  ioctl(audio_fd, SNDCTL_DSP_SPEED, &ao_data.samplerate);
}

/**
 * Stop playing, keep buffers (for pause).
 */
static void audio_pause() {
  uninit(1);
}

/**
 *  Resume playing, after audio_pause().
 */
static void audio_resume() {
  reset();
}

/**
 * Calculate how many bytes can be played without blocking, in full fragment
 * chunks.
 *
 * JJO: audio_buff_info.bytes is not used because it also includes bytes from 
 * partially empty fragments.
 */
static int get_space() {
  if ( ioctl(audio_fd, SNDCTL_DSP_GETOSPACE, &zz) != -1 ) {
    int playsize = zz.fragments * zz.fragsize;
    if (playsize > MAX_OUTBURST)
      playsize = (MAX_OUTBURST / zz.fragsize) * zz.fragsize;

    return playsize;
  }
  
  // JJO: Is this never executed.
#ifdef HAVE_AUDIO_SELECT
  {  fd_set rfds;
    struct timeval tv;
    
    mp_msg(MSGT_AO,MSGL_WARN,"[ttgo] Audio_getspace: Using select()\n");

    FD_ZERO(&rfds);
    FD_SET(audio_fd, &rfds);
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    if(!select(audio_fd+1, NULL, &rfds, NULL, &tv)) return 0; // not block!
  }
#endif

  return ao_data.outburst;
}

/**
 * Plays audio. 
 *
 * @param data Raw audio data.
 * @param len  Data size in bytes. 
 *
 * @return: number of bytes played.
 */
static int play(void* data,int len,int flags) {
  len = (len / ao_data.outburst) * ao_data.outburst;
  return write(audio_fd,data,len);
}

/**
 * Calculate delay in seconds between first and last sample in buffer.
 *
 * @return delay.
 */
static float get_delay() {
  if( ioctl(audio_fd, SNDCTL_DSP_GETOSPACE, &zz) != -1 )
    return ((float)(ao_data.buffersize-zz.bytes))/(float)ao_data.bps;
  else
    return ((float)ao_data.buffersize)/(float)ao_data.bps;
}
